#pragma once
#include <fstream>
#include <iostream>
#include <sstream>
#include <string>
#include <type_traits>
using namespace std;

// check if type is defined, forward declarations are not definitions
template <class, class = void>
struct is_defined : std::false_type {};
template <class T>
struct is_defined<
	T, std::enable_if_t<std::is_object<T>::value &&
	!std::is_pointer<T>::value && (sizeof(T) > 0)>>
	: std::true_type {};


namespace Test1 {

	template <class, class = void>
	struct is_defined_method : std::false_type {};

	template <class Message>
	struct is_defined_method<
		Message,
		std::enable_if_t<
		std::is_same<decltype(Message(std::declval<string>(), std::declval<int>())), Message>::value&&
		std::is_same<decltype(std::declval<std::ostream&>() << std::declval<const Message&>()), std::ostream&>::value
		>
	> : std::true_type {};

	template <class Message>
	int tst() {
		if constexpr (is_defined<Message>::value && is_defined_method<Message>::value) {
			std::stringstream ss;

			const Message msg("Ovo je poruka", 1);
			ss << msg;

			if (ss.str() == "Ovo je poruka") {
				std::cout << "Test passed" << std::endl;
				return 3;
			}
			else {
				std::cout << "Test failed" << std::endl;
			}
		}
		else {
			std::cout << "Interface failed" << std::endl;
		}
		return 0;
	}
}

namespace Test2 {

	template <class, class = void>
	struct is_defined_method : std::false_type {};

	template <class Message>
	struct is_defined_method<
		Message,
		std::enable_if_t<
		std::is_same<decltype(Message(std::declval<string>(), std::declval<int>())), Message>::value&&
		std::is_same<decltype(std::declval<Message>().setText(std::declval<string>())), void>::value&&
		std::is_same<decltype(std::declval<std::ostream&>() << std::declval<const Message&>()), std::ostream&>::value
		>
	> : std::true_type {};

	template <class Message>
	int tst() {
		if constexpr (is_defined<Message>::value && is_defined_method<Message>::value) {
			std::stringstream ss;

			Message msg("Ovo je poruka", 1);
			msg.setText("Ovo je nova poruka");
			ss << msg;

			if (ss.str() == "Ovo je nova poruka") {
				std::cout << "Test passed" << std::endl;
				return 2;
			}
			else {
				std::cout << "Test failed" << std::endl;
			}
		}
		else {
			std::cout << "Interface failed" << std::endl;
		}
		return 0;
	}
}

namespace Test3 {

	template <class, class = void>
	struct is_defined_method : std::false_type {};

	template <class Message>
	struct is_defined_method<
		Message,
		std::enable_if_t<
		std::is_same<decltype(Message(std::declval<string>(), std::declval<int>())), Message>::value&&
		std::is_same<decltype(std::declval<Message>().getSize()), int>::value&&
		std::is_same<decltype(std::declval<std::ostream&>() << std::declval<const Message&>()), std::ostream&>::value
		>
	> : std::true_type {};

	template <class Message>
	int tst() {
		if constexpr (is_defined<Message>::value && is_defined_method<Message>::value) {
			std::stringstream ss;

			Message msg("Zdravo svima!", 1);
			ss << msg.getSize();

			if (ss.str() == "13") {
				std::cout << "Test passed" << std::endl;
				return 2;
			}
			else {
				std::cout << "Test failed" << std::endl;
			}
		}
		else {
			std::cout << "Interface failed" << std::endl;
		}
		return 0;
	}
}

namespace Test4 {

	template <class, class = void>
	struct is_defined_method : std::false_type {};

	template <class Message>
	struct is_defined_method<
		Message,
		std::enable_if_t<
		std::is_same<decltype(Message(std::declval<string>(), std::declval<int>())), Message>::value&&
		std::is_same<decltype(std::declval<Message>() > std::declval<Message>()), bool>::value&&
		std::is_same<decltype(std::declval<std::ostream&>() << std::declval<const Message&>()), std::ostream&>::value
		>
	> : std::true_type {};

	template <class Message>
	int tst() {
		if constexpr (is_defined<Message>::value && is_defined_method<Message>::value) {
			std::stringstream ss;

			Message msg1("Ovo je poruka", 1);
			Message msg2("Ovo je poruka", 4);
			ss << (msg1 > msg2 ? "tacno" : "netacno");
			ss << (msg2 > msg1 ? "tacno" : "netacno");

			if (ss.str() == "tacnonetacno") {
				std::cout << "Test passed" << std::endl;
				return 3;
			}
			else {
				std::cout << "Test failed" << std::endl;
			}
		}
		else {
			std::cout << "Interface failed" << std::endl;
		}
		return 0;
	}
}



namespace Test5 {

	template <class, class, class = void>
	struct is_defined_method : std::false_type {};

	template <class Message, class TransmissionLine>
	struct is_defined_method<
		Message, TransmissionLine,
		std::enable_if_t<
		std::is_same<decltype(Message(std::declval<string>(), std::declval<int>())), Message>::value&&
		std::is_same<decltype(std::declval<std::ostream&>() << std::declval<const Message&>()), std::ostream&>::value&&
		std::is_same<decltype(TransmissionLine(std::declval<int>())), TransmissionLine>::value&&
		std::is_same<decltype(std::declval<TransmissionLine>().add(std::declval<Message*>())), void>::value&&
		std::is_same<decltype(std::declval<TransmissionLine>().pop()), Message*>::value
		>
	> : std::true_type {};

	template <class Message, class TransmissionLine>
	int tst() {
		if constexpr (is_defined<Message>::value && is_defined<TransmissionLine>::value && is_defined_method<Message, TransmissionLine>::value) {
			std::stringstream ss;

			Message msg1("Cao svima", 2);
			Message msg2("Zdravo svima", 5);
			TransmissionLine line(150);
			line.add(&msg1);
			line.add(&msg2);
			ss << *line.pop() << *line.pop();
			if (ss.str() == "Cao svimaZdravo svima") {
				std::cout << "Test passed" << std::endl;
				return 3;
			}
			else {
				std::cout << "Test failed" << std::endl;
			}
		}
		else {
			std::cout << "Interface failed" << std::endl;
		}
		return 0;
	}
}

namespace Test6 {

	template <class, class, class = void>
	struct is_defined_method : std::false_type {};

	template <class Message, class TransmissionLine>
	struct is_defined_method<
		Message, TransmissionLine,
		std::enable_if_t<
		std::is_same<decltype(Message(std::declval<string>(), std::declval<int>())), Message>::value&&
		std::is_same<decltype(std::declval<std::ostream&>() << std::declval<const Message&>()), std::ostream&>::value&&
		std::is_same<decltype(TransmissionLine(std::declval<int>())), TransmissionLine>::value&&
		std::is_same<decltype(std::declval<TransmissionLine>().add(std::declval<Message*>())), void>::value&&
		std::is_same<decltype(std::declval<TransmissionLine>().pop()), Message*>::value
		>
	> : std::true_type {};

	template <class Message, class TransmissionLine>
	int tst() {
		if constexpr (is_defined<Message>::value && is_defined<TransmissionLine>::value && is_defined_method<Message, TransmissionLine>::value) {
			std::stringstream ss;

			Message msg1("Cao svima", 7);
			Message msg2("Zdravo svima", 2);
			Message msg3("Pozdrav svima", 7);
			TransmissionLine line(150);
			line.add(&msg1);
			line.add(&msg2);
			line.add(&msg3);
			ss << *line.pop() << *line.pop() << *line.pop();
			if (ss.str() == "Zdravo svimaCao svimaPozdrav svima") {
				std::cout << "Test passed" << std::endl;
				return 3;
			}
			else {
				std::cout << "Test failed" << std::endl;
			}
		}
		else {
			std::cout << "Interface failed" << std::endl;
		}
		return 0;
	}
}

namespace Test7 {

	template <class, class, class = void>
	struct is_defined_method : std::false_type {};

	template <class Message, class TransmissionLine>
	struct is_defined_method<
		Message, TransmissionLine,
		std::enable_if_t<
		std::is_same<decltype(Message(std::declval<string>(), std::declval<int>())), Message>::value&&
		std::is_same<decltype(TransmissionLine(std::declval<int>())), TransmissionLine>::value&&
		std::is_same<decltype(std::declval<TransmissionLine>().add(std::declval<Message*>())), void>::value
		>
	> : std::true_type {};

	template <class Message, class TransmissionLine, class OverloadException>
	int tst() {
		if constexpr (is_defined<Message>::value && is_defined<TransmissionLine>::value && is_defined<OverloadException>::value && is_defined_method<Message, TransmissionLine>::value) {
			std::stringstream ss;

			try {
				Message msg1("Cao svima", 7);
				Message msg2("Zdravo svima", 2);
				Message msg3("Pozdrav svima", 6);
				TransmissionLine line(20);
				line.add(&msg1);
				line.add(&msg2);
				line.add(&msg3);
				std::cout << "Test failed" << std::endl;
			}
			catch (const OverloadException& e) {
				std::cout << "Test passed" << std::endl;
				return 2;
			}
		}
		else {
			std::cout << "Interface failed" << std::endl;
		}
		return 0;
	}
}

namespace Test8 {

	template <class, class, class = void>
	struct is_defined_method : std::false_type {};

	template <class Message, class TransmissionLine>
	struct is_defined_method<
		Message, TransmissionLine,
		std::enable_if_t<
		std::is_same<decltype(Message(std::declval<string>(), std::declval<int>())), Message>::value&&
		std::is_same<decltype(std::declval<std::ostream&>() << std::declval<const Message&>()), std::ostream&>::value&&
		std::is_same<decltype(TransmissionLine(std::declval<int>())), TransmissionLine>::value&&
		std::is_same<decltype(std::declval<TransmissionLine>().add(std::declval<Message*>())), void>::value&&
		std::is_same<decltype(std::declval<TransmissionLine>().pop()), Message*>::value
		>
	> : std::true_type {};

	template <class Message, class TransmissionLine, class LineEmptyException>
	int tst() {
		if constexpr (is_defined<Message>::value && is_defined<TransmissionLine>::value && is_defined<LineEmptyException>::value && is_defined_method<Message, TransmissionLine>::value) {
			std::stringstream ss;

			try {
				Message msg1("Cao svima", 7);
				Message msg2("Zdravo svima", 2);
				TransmissionLine line(200);
				line.add(&msg1);
				line.add(&msg2);
				line.pop(); line.pop(); line.pop();
				std::cout << "Test failed" << std::endl;
			}
			catch (const LineEmptyException& e) {
				std::cout << "Test passed" << std::endl;
				return 2;
			}
		}
		else {
			std::cout << "Interface failed" << std::endl;
		}
		return 0;
	}
}



namespace Test9 {

	template <class, class, class = void>
	struct is_defined_method : std::false_type {};

	template <class Cipher, class RowTranspositionCipher>
	struct is_defined_method<
		Cipher, RowTranspositionCipher,
		std::enable_if_t<
		std::is_same<decltype(RowTranspositionCipher(std::declval<int>())), RowTranspositionCipher>::value&&
		std::is_same<decltype(std::declval<Cipher>().encrypt(std::declval<string>())), string>::value&&
		std::is_same<decltype(std::declval<RowTranspositionCipher>().encrypt(std::declval<string>())), string>::value
		>
	> : std::true_type {};

	template <class Cipher, class RowTranspositionCipher>
	int tst() {
		if constexpr (is_defined<Cipher>::value && is_defined<RowTranspositionCipher>::value && is_defined_method<Cipher, RowTranspositionCipher>::value) {
			std::stringstream ss;

			Cipher* cipher = new RowTranspositionCipher(5);
			ss << cipher->encrypt("Cao svima koliko vas ima?");
			if (ss.str() == "Cvko aio iomlvm aiaa s ks?") {
				std::cout << "Test passed" << std::endl;
				return 3;
			}
			else {
				std::cout << "Test failed" << std::endl;
			}
		}
		else {
			std::cout << "Interface failed" << std::endl;
		}
		return 0;
	}
}

namespace Test10 {

	template <class, class, class = void>
	struct is_defined_method : std::false_type {};

	template <class Cipher, class RowTranspositionCipher>
	struct is_defined_method<
		Cipher, RowTranspositionCipher,
		std::enable_if_t<
		std::is_same<decltype(RowTranspositionCipher(std::declval<int>())), RowTranspositionCipher>::value&&
		std::is_same<decltype(std::declval<Cipher>().decrypt(std::declval<string>())), string>::value&&
		std::is_same<decltype(std::declval<RowTranspositionCipher>().decrypt(std::declval<string>())), string>::value
		>
	> : std::true_type {};

	template <class Cipher, class RowTranspositionCipher>
	int tst() {
		if constexpr (is_defined<Cipher>::value && is_defined<RowTranspositionCipher>::value && is_defined_method<Cipher, RowTranspositionCipher>::value) {
			std::stringstream ss;

			Cipher* cipher = new RowTranspositionCipher(6);
			ss << cipher->decrypt("Uofs rpdoeevssanin");
			if (ss.str() == "Uspesno desifrovan") {
				std::cout << "Test passed" << std::endl;
				return 2;
			}
			else {
				std::cout << "Test failed" << std::endl;
			}
		}
		else {
			std::cout << "Interface failed" << std::endl;
		}
		return 0;
	}
}

namespace Test11 {

	template <class, class, class = void>
	struct is_defined_method : std::false_type {};

	template <class Cipher, class RailFenceCipher>
	struct is_defined_method<
		Cipher, RailFenceCipher,
		std::enable_if_t<
		std::is_same<decltype(RailFenceCipher(std::declval<int>())), RailFenceCipher>::value&&
		std::is_same<decltype(std::declval<Cipher>().encrypt(std::declval<string>())), string>::value&&
		std::is_same<decltype(std::declval<RailFenceCipher>().encrypt(std::declval<string>())), string>::value
		>
	> : std::true_type {};

	template <class Cipher, class RailFenceCipher>
	int tst() {
		if constexpr (is_defined<Cipher>::value && is_defined<RailFenceCipher>::value && is_defined_method<Cipher, RailFenceCipher>::value) {
			std::stringstream ss;

			Cipher* cipher = new RailFenceCipher(5);
			ss << cipher->encrypt("Cao svima koliko vas ima?");
			if (ss.str() == "Ca ?am ovaoikkam voisisl ") {
				std::cout << "Test passed" << std::endl;
				return 3;
			}
			else {
				std::cout << "Test failed" << std::endl;
			}
		}
		else {
			std::cout << "Interface failed" << std::endl;
		}
		return 0;
	}
}

namespace Test12 {

	template <class, class, class = void>
	struct is_defined_method : std::false_type {};

	template <class Cipher, class RailFenceCipher>
	struct is_defined_method<
		Cipher, RailFenceCipher,
		std::enable_if_t<
		std::is_same<decltype(RailFenceCipher(std::declval<int>())), RailFenceCipher>::value&&
		std::is_same<decltype(std::declval<Cipher>().decrypt(std::declval<string>())), string>::value&&
		std::is_same<decltype(std::declval<RailFenceCipher>().decrypt(std::declval<string>())), string>::value
		>
	> : std::true_type {};

	template <class Cipher, class RailFenceCipher>
	int tst() {
		if constexpr (is_defined<Cipher>::value && is_defined<RailFenceCipher>::value && is_defined_method<Cipher, RailFenceCipher>::value) {
			std::stringstream ss;

			Cipher* cipher = new RailFenceCipher(2);
			ss << cipher->decrypt("Upsodsfoasen eirvn");
			if (ss.str() == "Uspesno desifrovan") {
				std::cout << "Test passed" << std::endl;
				return 2;
			}
			else {
				std::cout << "Test failed" << std::endl;
			}
		}
		else {
			std::cout << "Interface failed" << std::endl;
		}
		return 0;
	}
}



namespace Test13 {

	template <class, class, class, class, class, class, class = void>
	struct is_defined_method : std::false_type {};

	template <class Message, class TransmissionLine, class Cipher, class RowTranspositionCipher, class RailFenceCipher, class ProtectedTransmissionLine>
	struct is_defined_method<
		Message, TransmissionLine, Cipher, RowTranspositionCipher, RailFenceCipher, ProtectedTransmissionLine,
		std::enable_if_t<
		std::is_same<decltype(Message(std::declval<string>(), std::declval<int>())), Message>::value&&
		std::is_same<decltype(std::declval<std::ostream&>() << std::declval<const Message&>()), std::ostream&>::value&&
		std::is_same<decltype(TransmissionLine(std::declval<int>())), TransmissionLine>::value&&
		std::is_same<decltype(ProtectedTransmissionLine(std::declval<int>(), std::declval<Cipher*>())), ProtectedTransmissionLine>::value&&
		std::is_same<decltype(std::declval<TransmissionLine>().add(std::declval<Message*>())), void>::value&&
		std::is_same<decltype(std::declval<TransmissionLine>().pop()), Message*>::value&&
		std::is_same<decltype(std::declval<ProtectedTransmissionLine>().add(std::declval<Message*>())), void>::value&&
		std::is_same<decltype(std::declval<ProtectedTransmissionLine>().pop()), Message*>::value&&
		std::is_same<decltype(RowTranspositionCipher(std::declval<int>())), RowTranspositionCipher>::value
		>
	> : std::true_type {};

	template <class Message, class TransmissionLine, class Cipher, class RowTranspositionCipher, class RailFenceCipher, class ProtectedTransmissionLine>
	int tst() {
		if constexpr (is_defined<Message>::value && 
			is_defined<TransmissionLine>::value &&
			is_defined<Cipher>::value &&
			is_defined<RowTranspositionCipher>::value &&
			is_defined<RailFenceCipher>::value &&
			is_defined<ProtectedTransmissionLine>::value &&
			is_defined_method<Message, TransmissionLine, Cipher, RowTranspositionCipher, RailFenceCipher, ProtectedTransmissionLine>::value) {
			
			std::stringstream ss;

			Message msg1("Cao svima", 2);
			Message msg2("Zdravo svima", 5);
			Cipher* cipher = new RowTranspositionCipher(3);
			TransmissionLine* line = new ProtectedTransmissionLine(150, cipher);
			line->add(&msg1);
			line->add(&msg2);
			ss <<  *((ProtectedTransmissionLine*)line);
			if (ss.str() == "C iasmoua\nZa dvsmrova\n" || ss.str() == "C iasmoua\nZa dvsmrova") {
				std::cout << "Test passed" << std::endl;
				return 3;
			}
			else {
				std::cout << "Test failed" << std::endl;
			}
		}
		else {
			std::cout << "Interface failed" << std::endl;
		}
		return 0;
	}
}

namespace Test14 {

	template <class, class, class, class, class, class, class = void>
	struct is_defined_method : std::false_type {};

	template <class Message, class TransmissionLine, class Cipher, class RowTranspositionCipher, class RailFenceCipher, class ProtectedTransmissionLine>
	struct is_defined_method<
		Message, TransmissionLine, Cipher, RowTranspositionCipher, RailFenceCipher, ProtectedTransmissionLine,
		std::enable_if_t<
		std::is_same<decltype(Message(std::declval<string>(), std::declval<int>())), Message>::value&&
		std::is_same<decltype(std::declval<std::ostream&>() << std::declval<const Message&>()), std::ostream&>::value&&
		std::is_same<decltype(TransmissionLine(std::declval<int>())), TransmissionLine>::value&&
		std::is_same<decltype(ProtectedTransmissionLine(std::declval<int>(), std::declval<Cipher*>())), ProtectedTransmissionLine>::value&&
		std::is_same<decltype(std::declval<TransmissionLine>().add(std::declval<Message*>())), void>::value&&
		std::is_same<decltype(std::declval<TransmissionLine>().pop()), Message*>::value&&
		std::is_same<decltype(std::declval<ProtectedTransmissionLine>().add(std::declval<Message*>())), void>::value&&
		std::is_same<decltype(std::declval<ProtectedTransmissionLine>().pop()), Message*>::value&&
		std::is_same<decltype(RowTranspositionCipher(std::declval<int>())), RowTranspositionCipher>::value
		>
	> : std::true_type{};

	template <class Message, class TransmissionLine, class Cipher, class RowTranspositionCipher, class RailFenceCipher, class ProtectedTransmissionLine>
	int tst() {
		if constexpr (is_defined<Message>::value &&
			is_defined<TransmissionLine>::value &&
			is_defined<Cipher>::value &&
			is_defined<RowTranspositionCipher>::value &&
			is_defined<RailFenceCipher>::value &&
			is_defined<ProtectedTransmissionLine>::value &&
			is_defined_method<Message, TransmissionLine, Cipher, RowTranspositionCipher, RailFenceCipher, ProtectedTransmissionLine>::value) {

			std::stringstream ss;

			Message msg1("Cao svima", 2);
			Message msg2("Zdravo svima", 5);
			Cipher* cipher = new RowTranspositionCipher(3);
			TransmissionLine* line = new ProtectedTransmissionLine(150, cipher);
			line->add(&msg1);
			line->add(&msg2);
			ss << line->pop() << *((ProtectedTransmissionLine*)line);
			if (ss.str() == "Cao svimaZa dvsmrova\n" || ss.str() == "Cao svimaZa dvsmrova") {
				std::cout << "Test passed" << std::endl;
				return 2;
			}
			else {
				std::cout << "Test failed" << std::endl;
			}
		}
		else {
			std::cout << "Interface failed" << std::endl;
		}
		return 0;
	}
}